/*
 *   Copyright (C) 2009 Edward "Hades" Toroshchin <kde@hades.name>
 *   Copyright (C) 2008 Sun Microsystems, Inc.
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation, either version 3 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details
 *
 *   You should have received a copy of the GNU General Public
 *   License along with this program; if not, write to the
 *   Free Software Foundation, Inc.,
 *   51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 */

#include <QDomDocument>
#include <QFile>
#include <QFileInfo>
#include <QHash>
#include <QMutexLocker>

#include <KDebug>
#include <KIcon>
#include <KRun>

#include "vbox.h"

struct VBoxMachine
{
    QString name;
    QIcon   icon;
};

class VBoxConfigReader
{
    QList< VBoxMachine > *list;
    QDateTime lastChecked;
    QMutex mutex;

    QString vboxdir;
    QHash< QString, KIcon > mOsTypeIcons;
public:
    VBoxConfigReader()
     : list( 0 )
    {
        vboxdir = QString( getenv( "HOME" ) ) + "/.VirtualBox/";

        /* Fill in OS type icon dictionary */
        static const char *kOSTypeIcons [][2] =
        {
            {"Other",           "vbox-runner/os_other"},
            {"DOS",             "vbox-runner/os_dos"},
            {"Netware",         "vbox-runner/os_netware"},
            {"L4",              "vbox-runner/os_l4"},
            {"Windows31",       "vbox-runner/os_win31"},
            {"Windows95",       "vbox-runner/os_win95"},
            {"Windows98",       "vbox-runner/os_win98"},
            {"WindowsMe",       "vbox-runner/os_winme"},
            {"WindowsNT4",      "vbox-runner/os_winnt4"},
            {"Windows2000",     "vbox-runner/os_win2k"},
            {"WindowsXP",       "vbox-runner/os_winxp"},
            {"WindowsXP_64",    "vbox-runner/os_winxp_64"},
            {"Windows2003",     "vbox-runner/os_win2k3"},
            {"Windows2003_64",  "vbox-runner/os_win2k3_64"},
            {"WindowsVista",    "vbox-runner/os_winvista"},
            {"WindowsVista_64", "vbox-runner/os_winvista_64"},
            {"Windows2008",     "vbox-runner/os_win2k8"},
            {"Windows2008_64",  "vbox-runner/os_win2k8_64"},
            {"Windows7",        "vbox-runner/os_win7"},
            {"Windows7_64",     "vbox-runner/os_win7_64"},
            {"WindowsNT",       "vbox-runner/os_win_other"},
            {"OS2Warp3",        "vbox-runner/os_os2warp3"},
            {"OS2Warp4",        "vbox-runner/os_os2warp4"},
            {"OS2Warp45",       "vbox-runner/os_os2warp45"},
            {"OS2eCS",          "vbox-runner/os_os2ecs"},
            {"OS2",             "vbox-runner/os_os2_other"},
            {"Linux22",         "vbox-runner/os_linux22"},
            {"Linux24",         "vbox-runner/os_linux24"},
            {"Linux24_64",      "vbox-runner/os_linux24_64"},
            {"Linux26",         "vbox-runner/os_linux26"},
            {"Linux26_64",      "vbox-runner/os_linux26_64"},
            {"ArchLinux",       "vbox-runner/os_archlinux"},
            {"ArchLinux_64",    "vbox-runner/os_archlinux_64"},
            {"Debian",          "vbox-runner/os_debian"},
            {"Debian_64",       "vbox-runner/os_debian_64"},
            {"OpenSUSE",        "vbox-runner/os_opensuse"},
            {"OpenSUSE_64",     "vbox-runner/os_opensuse_64"},
            {"Fedora",          "vbox-runner/os_fedora"},
            {"Fedora_64",       "vbox-runner/os_fedora_64"},
            {"Gentoo",          "vbox-runner/os_gentoo"},
            {"Gentoo_64",       "vbox-runner/os_gentoo_64"},
            {"Mandriva",        "vbox-runner/os_mandriva"},
            {"Mandriva_64",     "vbox-runner/os_mandriva_64"},
            {"RedHat",          "vbox-runner/os_redhat"},
            {"RedHat_64",       "vbox-runner/os_redhat_64"},
            {"Ubuntu",          "vbox-runner/os_ubuntu"},
            {"Ubuntu_64",       "vbox-runner/os_ubuntu_64"},
            {"Xandros",         "vbox-runner/os_xandros"},
            {"Xandros_64",      "vbox-runner/os_xandros_64"},
            {"Linux",           "vbox-runner/os_linux_other"},
            {"FreeBSD",         "vbox-runner/os_freebsd"},
            {"FreeBSD_64",      "vbox-runner/os_freebsd_64"},
            {"OpenBSD",         "vbox-runner/os_openbsd"},
            {"OpenBSD_64",      "vbox-runner/os_openbsd_64"},
            {"NetBSD",          "vbox-runner/os_netbsd"},
            {"NetBSD_64",       "vbox-runner/os_netbsd_64"},
            {"Solaris",         "vbox-runner/os_solaris"},
            {"Solaris_64",      "vbox-runner/os_solaris_64"},
            {"OpenSolaris",     "vbox-runner/os_opensolaris"},
            {"OpenSolaris_64",  "vbox-runner/os_opensolaris_64"},
            {"QNX",             "vbox-runner/os_qnx"},
        };
        for (uint n = 0; n < sizeof (kOSTypeIcons) / sizeof( kOSTypeIcons[0] ); ++ n)
        {
            mOsTypeIcons.insert (kOSTypeIcons [n][0],
                KIcon (kOSTypeIcons [n][1]));
        }
    }

    ~VBoxConfigReader()
    {
        if( list ) delete list;
        list = 0;
    }

    void updateAsNeccessary()
    {
        QMutexLocker _ml( &mutex );
        if( !list || lastChecked < QFileInfo( vboxdir + "VirtualBox.xml" ).lastModified() )
        {
            if( list ) delete list;
            list = new QList< VBoxMachine >;

            QFile f( vboxdir + "VirtualBox.xml" );

            QDomDocument d;
            if( !d.setContent( &f ) )
            {
                kWarning() << "Could not parse file " << f.fileName();
                return;
            }

            QDomNodeList _dnlist = d.elementsByTagName( "MachineEntry" );
            for( int i = 0; i < _dnlist.count(); i++ )
            {
                QDomNode node = _dnlist.at( i );
                QString ref = node.toElement().attribute( "src" );
                if( ref.isEmpty() )
                {
                    kWarning() << "MachineEntry with no src attribute";
                    continue;
                }

                QFile m( vboxdir + ref );

                QDomDocument mspec;
                if( !mspec.setContent( &m ) )
                {
                    kWarning() << "Could not parse machine file " << m.fileName();
                    continue;
                }

                QDomNodeList _mlist = mspec.elementsByTagName( "Machine" );
                for( int j = 0; j < _mlist.count(); j++ )
                {
                    QDomNode mnode = _mlist.at( j );
                    VBoxMachine machine;
                    
                    machine.name = mnode.toElement().attribute( "name" );

                    QString type = mnode.toElement().attribute( "OSType" );
                    machine.icon = mOsTypeIcons.contains( type )?
                        mOsTypeIcons[ type ] : KIcon( "virtualbox-ose" );

                    (*list) << machine;
                }
            }

            lastChecked = QDateTime::currentDateTime();
        }
    }

    QList< VBoxMachine > machines()
    {
        updateAsNeccessary();
        if( !list )
            return QList< VBoxMachine >();

        return *list;
    }
};

VBoxRunner::VBoxRunner( QObject *parent, const QVariantList& args )
    : Plasma::AbstractRunner( parent, args ),
      rd( 0 )
{
    rd = new VBoxConfigReader;
    setObjectName( "VirtualBox Machines Runner" );
    setSpeed( AbstractRunner::SlowSpeed );
}

VBoxRunner::~VBoxRunner()
{
    if( rd ) delete rd;
    rd = 0;
}

void VBoxRunner::match(Plasma::RunnerContext &context)
{
    QList< Plasma::QueryMatch > matches;
    foreach( VBoxMachine m, rd->machines() )
        if( m.name.contains( context.query(), Qt::CaseInsensitive ) )
        {
            Plasma::QueryMatch match( this );
            match.setType( Plasma::QueryMatch::PossibleMatch );
            match.setIcon( m.icon );
            match.setText( m.name);
            match.setSubtext( i18n( "VirtualBox virtual machine" ) );
            matches << match;
        }

    context.addMatches(context.query(), matches);
}

void VBoxRunner::run(const Plasma::RunnerContext &context, const Plasma::QueryMatch &match)
{
    Q_UNUSED(context)

    KRun::runCommand( QString( "VBoxManage startvm \"%1\"" ).arg( match.text() ), 0 );
}

